package org.apache.bcel.classfile;

package .var.folders.cl.9twcgjfx4vsg5zl73c6wxmvm0000gn.T;

import java.io.DataInput;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Arrays;
import java.util.Iterator;
import org.apache.bcel.Const;
import org.apache.bcel.classfile.ClassFormatException;
import org.apache.bcel.classfile.Constant;
import org.apache.bcel.classfile.ConstantCP;
import org.apache.bcel.classfile.ConstantClass;
import org.apache.bcel.classfile.ConstantDouble;
import org.apache.bcel.classfile.ConstantDynamic;
import org.apache.bcel.classfile.ConstantFloat;
import org.apache.bcel.classfile.ConstantInteger;
import org.apache.bcel.classfile.ConstantInvokeDynamic;
import org.apache.bcel.classfile.ConstantLong;
import org.apache.bcel.classfile.ConstantMethodHandle;
import org.apache.bcel.classfile.ConstantMethodType;
import org.apache.bcel.classfile.ConstantModule;
import org.apache.bcel.classfile.ConstantNameAndType;
import org.apache.bcel.classfile.ConstantPackage;
import org.apache.bcel.classfile.ConstantString;
import org.apache.bcel.classfile.ConstantUtf8;
import org.apache.bcel.classfile.Node;
import org.apache.bcel.classfile.Utility;
import org.apache.bcel.classfile.Visitor;

public class ConstantPool implements Cloneable, Node, Iterable<Constant> {
  private Constant[] constantPool;
  
  private static String escape(String str) {
    int len = str.length();
    StringBuilder buf = new StringBuilder(len + 5);
    char[] ch = str.toCharArray();
    for (int i = 0; i < len; i++) {
      switch (ch[i]) {
        case '\n':
          buf.append("\\n");
          break;
        case '\r':
          buf.append("\\r");
          break;
        case '\t':
          buf.append("\\t");
          break;
        case '\b':
          buf.append("\\b");
          break;
        case '"':
          buf.append("\\\"");
          break;
        default:
          buf.append(ch[i]);
          break;
      } 
    } 
    return buf.toString();
  }
  
  public ConstantPool(Constant[] constantPool) {
    this.constantPool = constantPool;
  }
  
  public ConstantPool(DataInput input) throws IOException {
    int constantPoolCount = input.readUnsignedShort();
    this.constantPool = new Constant[constantPoolCount];
    for (int i = 1; i < constantPoolCount; i++) {
      this.constantPool[i] = Constant.readConstant(input);
      byte tag = this.constantPool[i].getTag();
      if (tag == 6 || tag == 5)
        i++; 
    } 
  }
  
  public void accept(Visitor v) {
    v.visitConstantPool(this);
  }
  
  public String constantToString(Constant c) throws IllegalArgumentException {
    String str;
    int i;
    ConstantMethodHandle cmh;
    ConstantMethodType cmt;
    ConstantInvokeDynamic cid;
    ConstantDynamic cd;
    byte tag = c.getTag();
    switch (tag) {
      case 7:
        i = ((ConstantClass)c).getNameIndex();
        c = getConstantUtf8(i);
        str = Utility.compactClassName(((ConstantUtf8)c).getBytes(), false);
        return str;
      case 8:
        i = ((ConstantString)c).getStringIndex();
        c = getConstantUtf8(i);
        str = "\"" + escape(((ConstantUtf8)c).getBytes()) + "\"";
        return str;
      case 1:
        str = ((ConstantUtf8)c).getBytes();
        return str;
      case 6:
        str = String.valueOf(((ConstantDouble)c).getBytes());
        return str;
      case 4:
        str = String.valueOf(((ConstantFloat)c).getBytes());
        return str;
      case 5:
        str = String.valueOf(((ConstantLong)c).getBytes());
        return str;
      case 3:
        str = String.valueOf(((ConstantInteger)c).getBytes());
        return str;
      case 12:
        str = constantToString(((ConstantNameAndType)c).getNameIndex(), (byte)1) + " " + constantToString(((ConstantNameAndType)c).getSignatureIndex(), (byte)1);
        return str;
      case 9:
      case 10:
      case 11:
        str = constantToString(((ConstantCP)c).getClassIndex(), (byte)7) + "." + constantToString(((ConstantCP)c).getNameAndTypeIndex(), (byte)12);
        return str;
      case 15:
        cmh = (ConstantMethodHandle)c;
        str = Const.getMethodHandleName(cmh.getReferenceKind()) + " " + constantToString(cmh.getReferenceIndex(), getConstant(cmh.getReferenceIndex()).getTag());
        return str;
      case 16:
        cmt = (ConstantMethodType)c;
        str = constantToString(cmt.getDescriptorIndex(), (byte)1);
        return str;
      case 18:
        cid = (ConstantInvokeDynamic)c;
        str = cid.getBootstrapMethodAttrIndex() + ":" + constantToString(cid.getNameAndTypeIndex(), (byte)12);
        return str;
      case 17:
        cd = (ConstantDynamic)c;
        str = cd.getBootstrapMethodAttrIndex() + ":" + constantToString(cd.getNameAndTypeIndex(), (byte)12);
        return str;
      case 19:
        i = ((ConstantModule)c).getNameIndex();
        c = getConstantUtf8(i);
        str = Utility.compactClassName(((ConstantUtf8)c).getBytes(), false);
        return str;
      case 20:
        i = ((ConstantPackage)c).getNameIndex();
        c = getConstantUtf8(i);
        str = Utility.compactClassName(((ConstantUtf8)c).getBytes(), false);
        return str;
    } 
    throw new IllegalArgumentException("Unknown constant type " + tag);
  }
  
  public String constantToString(int index, byte tag) {
    return constantToString(getConstant(index, tag));
  }
  
  public org.apache.bcel.classfile.ConstantPool copy() {
    org.apache.bcel.classfile.ConstantPool c = null;
    try {
      c = (org.apache.bcel.classfile.ConstantPool)clone();
      c.constantPool = new Constant[this.constantPool.length];
      for (int i = 1; i < this.constantPool.length; i++) {
        if (this.constantPool[i] != null)
          c.constantPool[i] = this.constantPool[i].copy(); 
      } 
    } catch (CloneNotSupportedException cloneNotSupportedException) {}
    return c;
  }
  
  public void dump(DataOutputStream file) throws IOException {
    int size = Math.min(this.constantPool.length, 65535);
    file.writeShort(size);
    for (int i = 1; i < size; i++) {
      if (this.constantPool[i] != null)
        this.constantPool[i].dump(file); 
    } 
  }
  
  public <T extends Constant> T getConstant(int index) throws ClassFormatException {
    return getConstant(index, (Class)Constant.class);
  }
  
  public <T extends Constant> T getConstant(int index, byte tag) throws ClassFormatException {
    return getConstant(index, tag, (Class)Constant.class);
  }
  
  public <T extends Constant> T getConstant(int index, byte tag, Class<T> castTo) throws ClassFormatException {
    T c = getConstant(index);
    if (c.getTag() != tag)
      throw new ClassFormatException("Expected class `" + Const.getConstantName(tag) + "' at index " + index + " and got " + c); 
    return c;
  }
  
  public <T extends Constant> T getConstant(int index, Class<T> castTo) throws ClassFormatException {
    if (index >= this.constantPool.length || index < 0)
      throw new ClassFormatException("Invalid constant pool reference: " + index + ". Constant pool size is: " + this.constantPool.length); 
    Constant constant = (Constant)castTo.cast(this.constantPool[index]);
    if (constant == null && index != 0) {
      Constant prev = this.constantPool[index - 1];
      if (prev == null || (prev.getTag() != 6 && prev.getTag() != 5))
        throw new ClassFormatException("Constant pool at index " + index + " is null."); 
    } 
    return (T)constant;
  }
  
  public ConstantInteger getConstantInteger(int index) {
    return getConstant(index, (byte)3, ConstantInteger.class);
  }
  
  public Constant[] getConstantPool() {
    return this.constantPool;
  }
  
  public String getConstantString(int index, byte tag) throws IllegalArgumentException {
    int i;
    switch (tag) {
      case 7:
        i = ((ConstantClass)getConstant(index, ConstantClass.class)).getNameIndex();
        return getConstantUtf8(i).getBytes();
      case 8:
        i = ((ConstantString)getConstant(index, ConstantString.class)).getStringIndex();
        return getConstantUtf8(i).getBytes();
      case 19:
        i = ((ConstantModule)getConstant(index, ConstantModule.class)).getNameIndex();
        return getConstantUtf8(i).getBytes();
      case 20:
        i = ((ConstantPackage)getConstant(index, ConstantPackage.class)).getNameIndex();
        return getConstantUtf8(i).getBytes();
      case 1:
        return getConstantUtf8(index).getBytes();
    } 
    throw new IllegalArgumentException("getConstantString called with illegal tag " + tag);
  }
  
  public ConstantUtf8 getConstantUtf8(int index) throws ClassFormatException {
    return getConstant(index, (byte)1, ConstantUtf8.class);
  }
  
  public int getLength() {
    return (this.constantPool == null) ? 0 : this.constantPool.length;
  }
  
  public Iterator<Constant> iterator() {
    return Arrays.<Constant>stream(this.constantPool).iterator();
  }
  
  public void setConstant(int index, Constant constant) {
    this.constantPool[index] = constant;
  }
  
  public void setConstantPool(Constant[] constantPool) {
    this.constantPool = constantPool;
  }
  
  public String toString() {
    StringBuilder buf = new StringBuilder();
    for (int i = 1; i < this.constantPool.length; i++)
      buf.append(i).append(")").append(this.constantPool[i]).append("\n"); 
    return buf.toString();
  }
}
